/* really simple example of using libjpeg-6b with kallistios */

#include <stdio.h>
#include <kallisti/libk.h>
#include <kallisti/abi/fs.h>
#include <kallisti/abi/video.h>
#include <kallisti/abi/thread.h>
#include <kallisti/abi/maple.h>

#include "jpeglib.h"

/* Library handles */
abi_fs_t    *fs = NULL;
abi_thread_t    *thread = NULL;
abi_video_t *video = NULL;
abi_maple_t *maple = NULL;

unsigned int y = 0;

/* total crap, assumes a 640x480 or less jpeg */
void put_scanline_someplace(char *buffer, int row_stride)
{
    unsigned short ourbuffer[640];
    int i;

    for(i=0; i<row_stride/3; i++)
        ourbuffer[i] = ((buffer[i*3]>>3)<<11) + ((buffer[i*3+1]>>2)<<5) + (buffer[i*3+2]>>3);

    memcpy(video->vram_s + 640*y, ourbuffer, 2*row_stride/3); 
    
    y++;
}

int read_JPEG_file (char * filename)
{
    /* This struct contains the JPEG decompression parameters and pointers to
     * working space (which is allocated as needed by the JPEG library).
     */
    struct jpeg_decompress_struct cinfo;
    
    struct jpeg_error_mgr jerr;
    
    /* More stuff */
    FILE * infile;                /* source file */
    JSAMPARRAY buffer;            /* Output row buffer */
    int row_stride;               /* physical row width in output buffer */
    
    /* In this example we want to open the input file before doing anything else,
     * so that the setjmp() error recovery below can assume the file is open.
     * VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
     * requires it in order to read binary files.
     */
        
    if ((infile = fopen(filename, "rb")) == NULL) {
	fprintf(stderr, "can't open %s\n", filename);
	return 0;
    }
    
    /* Step 1: allocate and initialize JPEG decompression object */
    
    /* We set up the normal JPEG error routines */
    cinfo.err = jpeg_std_error(&jerr);
    
    /* Now we can initialize the JPEG decompression object. */
    jpeg_create_decompress(&cinfo);
    
    /* Step 2: specify data source (eg, a file) */
    
    jpeg_stdio_src(&cinfo, infile);
    
    /* Step 3: read file parameters with jpeg_read_header() */
    
    (void) jpeg_read_header(&cinfo, TRUE);
    /* We can ignore the return value from jpeg_read_header since
     *   (a) suspension is not possible with the stdio data source, and
     *   (b) we passed TRUE to reject a tables-only JPEG file as an error.
     * See libjpeg.doc for more info.
     */
    
    /* Step 4: set parameters for decompression */
    
    /* In this example, we don't need to change any of the defaults set by
     * jpeg_read_header(), so we do nothing here.
     */

//    cinfo.dct_method = JDCT_FLOAT;

    /* Step 5: Start decompressor */
    
    (void) jpeg_start_decompress(&cinfo);
    /* We can ignore the return value since suspension is not possible
     * with the stdio data source.
     */
    
    /* We may need to do some setup of our own at this point before reading
     * the data.  After jpeg_start_decompress() we have the correct scaled
     * output image dimensions available, as well as the output colormap
     * if we asked for color quantization.
     * In this example, we need to make an output work buffer of the right size.
     */ 
    /* JSAMPLEs per row in output buffer */
    row_stride = cinfo.output_width * cinfo.output_components;
    /* Make a one-row-high sample array that will go away when done with image */
    buffer = (*cinfo.mem->alloc_sarray)
	((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
    
    /* Step 6: while (scan lines remain to be read) */
    /*           jpeg_read_scanlines(...); */
    
    /* Here we use the library's state variable cinfo.output_scanline as the
     * loop counter, so that we don't have to keep track ourselves.
     */
    while (cinfo.output_scanline < cinfo.output_height) {
	/* jpeg_read_scanlines expects an array of pointers to scanlines.
	 * Here the array is only one element long, but you could ask for
	 * more than one scanline at a time if that's more convenient.
	 */
	(void) jpeg_read_scanlines(&cinfo, buffer, 1);
	/* Assume put_scanline_someplace wants a pointer and sample count. */
	put_scanline_someplace(buffer[0], row_stride);
    }
    
    /* Step 7: Finish decompression */
    
    (void) jpeg_finish_decompress(&cinfo);
    /* We can ignore the return value since suspension is not possible
     * with the stdio data source.
     */
    
    /* Step 8: Release JPEG decompression object */
    
    /* This is an important step since it will release a good deal of memory. */
    jpeg_destroy_decompress(&cinfo);
    
    /* After finish_decompress, we can close the input file.
     * Here we postpone it until after no more JPEG errors are possible,
     * so as to simplify the setjmp error logic above.  (Actually, I don't
     * think that jpeg_destroy can do an error exit, but why assume anything...)
     */
    fclose(infile);
    
    /* At this point you may want to check to see whether any corrupt-data
     * warnings occurred (test whether jerr.pub.num_warnings is nonzero).
     */
    
    /* And we're done! */
    return 1;
}

int main()
{
    unsigned int start, end;
    cont_cond_t cond;

    printf("Program starting\r\n");
    
    fs = lib_open("fs");
    if (fs == NULL) { printf("Can't open fs lib\r\n"); return -1; }
    video = lib_open("video");
    if (video == NULL) { printf("Can't open video lib\r\n"); return -1; }
    thread = lib_open("thread");
    if (thread == NULL) { printf("Can't open thread lib\r\n"); return -1; }
    maple = lib_open("maple");
    if (maple == NULL) { printf("Can't open maple lib\r\n"); return -1; }

    video->set_mode(video->check_cable(), DM_640x480, PM_RGB565);

    printf("Loading and displaying JPEG...");

    start = thread->jiffies();

    read_JPEG_file("/rd/test.jpg");

    end = thread->jiffies();

    printf("took %d jiffies\r\n",end-start);

    printf("Press start to exit\r\n");

    while(1) {
        if (maple->cont_get_cond(maple->first_controller(), &cond) < 0)
            break;
        
        if (!(cond.buttons & CONT_START))
            break;
	
	thread->sleep(10);
    }

    printf("Done, returning\r\n");
    return 0;
}
